/*
 * SuperVESA FrameBuffer daemon. Version 1.1
 *
 * (C) 2003 Raster Software Vigo (raster@rastersoft.com)
 *
 * http://www.rastersoft.com
 *
 * This program is launched from the SuperVESA FrameBuffer driver
 * each time it must change the graphics mode.
 *
 * It's implemented as a daemon instead of inserting the code into
 * the kernel because:
 *
 *   a) The LRMI library is designed to work in user space
 *   b) I see *extremely dangerous* to call BIOS' 16-bit code from
 *      inside the kernel
 *   c) I'm too lazy to even try to do it :)
 *
 *
 * This program is in the public domain (GPL)
 * and heavily based in VBETEST.C from LRMI library
 * (http://sourceforge.net/projects/lrmi)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#if defined(__linux__)
#include <sys/io.h>
#include <sys/kd.h>
#include <sys/stat.h>
#elif defined(__NetBSD__)
#include <time.h>
#include <dev/wscons/wsconsio.h>
#include <machine/sysarch.h>
#elif defined(__FreeBSD__)
#include <machine/console.h>
#include <machine/sysarch.h>
#endif

#include "lrmi-0.8/lrmi.h"
#include "lrmi-0.8/vbe.h"

#define dac_reg	(0x3c8)
#define dac_val	(0x3c9)

unsigned short int vesamodes[1542]; // up to 257 entries
unsigned char color_data[5][3][2];
int number; // number of entries

struct {
	struct vbe_info_block *info;
	struct vbe_mode_info_block *mode;
	char *win; 	/* this doesn't point directly at the window, see update_window() */
	int win_low, win_high;
} vbe;

int store_modes() { // store the found modes in /etc/supervesafb/modes.info
	
	FILE *salida;
	int bucle,bucle2;
	unsigned char *pointer_entry;
	unsigned char comando[6];
	
	// create the directory
	mkdir("/etc/supervesafb",448); // 448 -> 700 for perm (rwx------)
	
	salida=fopen("/etc/supervesafb/modes.info","w");
	if(salida==NULL) {
		printf("Can't create the file /ect/supervesafb/modes.info. Aborting\n");
		exit(1);
	}
	
	fprintf(salida,"%c%c******",0,(unsigned char)number); // start command
	
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,0,
		color_data[0][0][0],color_data[0][0][1],color_data[0][1][0],
		color_data[0][1][1],color_data[0][2][0],color_data[0][2][1]);
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,1,
		color_data[1][0][0],color_data[1][0][1],color_data[1][1][0],
		color_data[1][1][1],color_data[1][2][0],color_data[1][2][1]);
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,2,
		color_data[2][0][0],color_data[2][0][1],color_data[2][1][0],
		color_data[2][1][1],color_data[2][2][0],color_data[2][2][1]);
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,3,
		color_data[3][0][0],color_data[3][0][1],color_data[3][1][0],
		color_data[3][1][1],color_data[3][2][0],color_data[3][2][1]);
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,4,
		color_data[4][0][0],color_data[4][0][1],color_data[4][1][0],
		color_data[4][1][1],color_data[4][2][0],color_data[4][2][1]);	
	
	pointer_entry = (unsigned char *) vesamodes;
	
	bucle=0;
	do {
		for(bucle2=0;bucle2<6;bucle2++)
			comando[bucle2]=*(pointer_entry++);
		fprintf(salida,"%c%c%c%c%c%c%c%c",2,(unsigned char) bucle,comando[0],comando[1],comando[2],comando[3],comando[4],comando[5]);
		bucle++;
	} while ((comando[0]!=0)||(comando[1]!=0));
	
	fprintf(salida,"%c*******",3);

	fclose(salida);
	
	printf("FBSETMODED: %d graphic modes stored\n",bucle-1);
	
	return 0;
}
	

int send_current() {

	/* Here we send all the commands to the kernel, as found by FINDMODES.
		We open and close the /proc/supervesafb file each time to
		ensure that each command is sent individually to the kernel callback.
		It's not elegant, but this is done only once in each boot... */

	FILE *salida;
	int bucle,bucle2;
	unsigned char *pointer_entry;
	unsigned char comando[6];
	
	salida=fopen("/proc/supervesafb","w");
	if(salida==NULL) {
	  printf("Can't open /proc/supervesafb file for output. Aborting\n");
	  exit(1);
	}

	fprintf(salida,"%c%c******",0,(unsigned char)number); // start command
	
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
		
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,0,
		color_data[0][0][0],color_data[0][0][1],color_data[0][1][0],
		color_data[0][1][1],color_data[0][2][0],color_data[0][2][1]);
	
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
	
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,1,
		color_data[1][0][0],color_data[1][0][1],color_data[1][1][0],
		color_data[1][1][1],color_data[1][2][0],color_data[1][2][1]);
	
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
	
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,2,
		color_data[2][0][0],color_data[2][0][1],color_data[2][1][0],
		color_data[2][1][1],color_data[2][2][0],color_data[2][2][1]);
		
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
		
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,3,
		color_data[3][0][0],color_data[3][0][1],color_data[3][1][0],
		color_data[3][1][1],color_data[3][2][0],color_data[3][2][1]);
		
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
		
	fprintf(salida,"%c%c%c%c%c%c%c%c",1,4,
		color_data[4][0][0],color_data[4][0][1],color_data[4][1][0],
		color_data[4][1][1],color_data[4][2][0],color_data[4][2][1]);	
	
	pointer_entry = (unsigned char *) vesamodes;
	
	bucle=0;
	do {
		fclose(salida);
		salida=fopen("/proc/supervesafb","w");
		for(bucle2=0;bucle2<6;bucle2++)
			comando[bucle2]=*(pointer_entry++);
		fprintf(salida,"%c%c%c%c%c%c%c%c",2,(unsigned char) bucle,comando[0],comando[1],comando[2],comando[3],comando[4],comando[5]);
		bucle++;
	} while ((comando[0]!=0)||(comando[1]!=0));
	
	fclose(salida);
	salida=fopen("/proc/supervesafb","w");
	
	fprintf(salida,"%c*******",3);
	fclose(salida);
	
	printf("FBSETMODED: %d graphic modes found\n",bucle-1);
	
	return 0;
}

 /* tests if there's the file /etc/supervesafb/modes.info
	if it exists, sends its info to the driver and returns 0; if not, returns 1 */

int send_stored () {
	
	FILE *fichero,*salida;
	char c[8];
	int counter;
	
	fichero=fopen("/etc/supervesafb/modes.info","r");
	if(fichero==NULL)
		return 1; // the file doesn't exist
		
	counter=0;
	
	do {
		salida=fopen("/proc/supervesafb","w");
		if(salida==NULL) {
			printf("Can't open /proc/supervesafb for writting. Aborting\n");
			exit(1);
		}
		fscanf(fichero,"%c%c%c%c%c%c%c%c",&c[0],&c[1],&c[2],&c[3],&c[4],&c[5],&c[6],&c[7]);
		fprintf(salida,"%c%c%c%c%c%c%c%c",c[0],c[1],c[2],c[3],c[4],c[5],c[6],c[7]);
		if(c[0]==2)
			counter++; // it's a graphic mode command
		fclose(salida);
	} while (c[0]!=3); // end command
	
	printf("FBSETMODED: %d graphic modes found in MODES.INFO\n",counter-1);
	return 0;
}
		

int findmodes() { // finds the available graphic modes and stores them in VESAMODES and COLOR_DATA

	struct LRMI_regs r;
	short int *mode_list;

	unsigned short int *current_entry;
	// unsigned char *pointer_entry;
	unsigned short int resx,resy,bpp,mode;
	
#if defined(__NetBSD__)
	unsigned long iomap[32];
#endif


	// FILE *salida;
		
	//unsigned char comando[6];
	int bucle,bucle2,bucle3;
	int indice,hastodoit;

	current_entry = vesamodes;
	
	for(bucle=0;bucle<1542;bucle++)
	  vesamodes[bucle]=0;
	
	for(bucle=0;bucle<5;bucle++)
	  for(bucle2=0;bucle2<3;bucle2++)
	    for(bucle3=0;bucle3<2;bucle3++)
	      color_data[bucle][bucle2][bucle3]=0;

	vbe.info = LRMI_alloc_real(sizeof(struct vbe_info_block)
	 + sizeof(struct vbe_mode_info_block));

	if (vbe.info == NULL) {
		fprintf(stderr, "Can't alloc real mode memory\n");
		return 1;
	}

	vbe.mode = (struct vbe_mode_info_block *)(vbe.info + 1);

#if 0
	/*
	 Allow read/write to video IO ports
	*/
	ioperm(0x2b0, 0x2df - 0x2b0, 1);
	ioperm(0x3b0, 0x3df - 0x3b0, 1);
#else
	/*
	 Allow read/write to ALL io ports
	*/
#if defined(__linux__)
	ioperm(0, 1024, 1);
	iopl(3);
#elif defined(__NetBSD__)
	memset(&iomap[0], 0xff, sizeof(iomap));
	i386_set_ioperm(iomap);
	i386_iopl(3);
#elif defined(__FreeBSD__)
	i386_set_ioperm(0, 0x10000, 1);
#endif
#endif

	memset(&r, 0, sizeof(r));

	r.eax = 0x4f00;
	r.es = (unsigned int)vbe.info >> 4;
	r.edi = 0;

	memcpy(vbe.info->vbe_signature, "VBE2", 4);

	if (!LRMI_int(0x10, &r)) {
		fprintf(stderr, "Can't get VESA info (vm86 failure)\n");
		return 1;
	}

	if ((r.eax & 0xffff) != 0x4f || strncmp(vbe.info->vbe_signature, "VESA", 4) != 0) {
		fprintf(stderr, "No VESA bios\n");
		return 1;
	}

	/* printf("VBE Version %x.%x\n",
	 (int)(vbe.info->vbe_version >> 8) & 0xff,
	 (int)vbe.info->vbe_version & 0xff);

	printf("%s\n",
	 (char *)(vbe.info->oem_string_seg * 16 + vbe.info->oem_string_off)); */

	mode_list = (short int *)(vbe.info->video_mode_list_seg * 16 + vbe.info->video_mode_list_off);

	number=0;
	
	while (*mode_list != -1) {
		memset(&r, 0, sizeof(r));

		r.eax = 0x4f01;
		r.ecx = *mode_list;
		r.es = (unsigned int)vbe.mode >> 4;
		r.edi = (unsigned int)vbe.mode & 0xf;

		if (!LRMI_int(0x10, &r)) {
			fprintf(stderr, "Can't get mode info (vm86 failure)\n");
			return 1;
		}

		hastodoit=0;
		
		if ((vbe.mode->memory_model == VBE_MODEL_RGB)&&((*mode_list)>255)) {

		  indice=0;

		  if(vbe.mode->bits_per_pixel==32)
		    indice=4;
		  if(vbe.mode->bits_per_pixel==24)
			indice=3;
		  if(vbe.mode->bits_per_pixel==16)
		    indice=2;
		  if(vbe.mode->bits_per_pixel==15)
		    indice=1;

		  if(indice!=0)
			  hastodoit=1;

		} else if (((vbe.mode->memory_model == VBE_MODEL_256)||(vbe.mode->memory_model == VBE_MODEL_PACKED))&&((*mode_list)>255)) {

			if(vbe.mode->bits_per_pixel == 8) {
		  		indice=0;
				hastodoit=2;
			}
		}
		
		if(hastodoit!=0) {

			number++;

			if(hastodoit==1) {
				if(vbe.mode->red_mask_size!=0) {
					color_data[indice][0][0]=vbe.mode->red_mask_size;
					color_data[indice][0][1]=vbe.mode->red_field_position;
				}
				if(vbe.mode->green_mask_size!=0) {
					color_data[indice][1][0]=vbe.mode->green_mask_size;
					color_data[indice][1][1]=vbe.mode->green_field_position;		    
				}
				if(vbe.mode->blue_mask_size!=0) {
					color_data[indice][2][0]=vbe.mode->blue_mask_size;
					color_data[indice][2][1]=vbe.mode->blue_field_position;
				}
			} else {
			  	color_data[indice][0][0]=vbe.mode->red_mask_size ? vbe.mode->red_mask_size : 8;
		  		color_data[indice][1][0]=vbe.mode->green_mask_size ? vbe.mode->green_mask_size : 8;
		  		color_data[indice][2][0]=vbe.mode->blue_mask_size ? vbe.mode->blue_mask_size : 8;				
			}
			
		    resx = (unsigned short int) vbe.mode->x_resolution;
			resy = (unsigned short int) vbe.mode->y_resolution;
			bpp = (unsigned short int) vbe.mode->bits_per_pixel;
			mode= (unsigned short int) (*mode_list);
			  
			*current_entry = resx;
			current_entry++;
			*current_entry = resy;
			current_entry++;
			*current_entry = (((bpp<<8) & 0xFF00) | (mode & 0x00FF));
			current_entry++;
		}

		mode_list++;
	}

	LRMI_free_real(vbe.info);

	return 0;
	
}
	

void
set_mode(int n)
{
	struct LRMI_regs r;

	memset(&r, 0, sizeof(r));

	r.eax = 0x4f02;
	r.ebx = n;

	if (!LRMI_int(0x10, &r)) {
		fprintf(stderr, "Can't set video mode (vm86 failure)\n");
	} else if ((r.eax & 0xffff) != 0x4f) {
		fprintf(stderr, "Set video mode failed\n");
	}

}

void cambia_modo(unsigned char nuevo_modo) {
	
	struct LRMI_regs r;
	short int modo;
	
	// FILE *debug_file;
	
#if defined(__NetBSD__)
	unsigned long iomap[32];
#endif
	
	vbe.info = LRMI_alloc_real(sizeof(struct vbe_info_block)
								+ sizeof(struct vbe_mode_info_block));

	if (vbe.info == NULL) {
		fprintf(stderr, "Can't alloc real mode memory\n");
		return;
	}

	vbe.mode = (struct vbe_mode_info_block *)(vbe.info + 1);

	memset(&r, 0, sizeof(r));

	r.eax = 0x4f00;
	r.es = (unsigned int)vbe.info >> 4;
	r.edi = 0;

	memcpy(vbe.info->vbe_signature, "VBE2", 4);
	
	iopl(3);
	
	if (!LRMI_int(0x10, &r)) {
		fprintf(stderr, "Can't get VESA info (vm86 failure)\n");
		return;
	}

	if ((r.eax & 0xffff) != 0x4f || strncmp(vbe.info->vbe_signature, "VESA", 4) != 0) {
		fprintf(stderr, "No VESA bios\n");
		return;
	}

	LRMI_free_real(vbe.info);
	
	modo=((short int) nuevo_modo)|0xC100; // 4 linear framebuffer && 8 don't erase video mem && 1 for VESA modes
	
	set_mode(modo);
	
	return;
}

int main(int argc, char *argv[])
{
	int maxconsole;
	int pid_value;

	FILE *options,*salida;
	unsigned char buffer[8]; // here we store the command sent by the kernel

	if (!LRMI_init())
		return 1;

	if((argc == 2) && (0 == strcasecmp(argv[1],"findmodes"))) { // finds the graphic modes
		if(0==send_stored())
			return 0; // sent stored modes
		if(0==findmodes()) // no stored modes. Find and send them
			return(send_current());
		else
			return (-1);
	}
	
	if((argc == 2) && (0 == strcasecmp(argv[1],"storemodes"))) { // stores the graphic modes
		if(0==findmodes())
			store_modes();
		return 0;
	}
	
	if(argc != 1) {
		printf("\nInvalid command.\nUsage: fbsetmoded [ findmodes | storemodes ]\n\n");
		exit(1);
	}
	
	// read the X-console number
	
	options=fopen("/etc/supervesafb/xconsole","r");
	if(options==NULL)
		maxconsole=6; // X are in console 6 (Alt+F7)
	else {
		fscanf(options,"%d",&maxconsole);
		fclose(options);
	}
	
	if((maxconsole<1)||(maxconsole>23))
		maxconsole=6;
	
	// open the /proc file
	
	salida=fopen("/proc/supervesafb","r"); // open the /proc file for read
	if(salida==NULL) { // there's no SUPERVESAFB driver!!!
		printf("There's no SuperVESAfb driver compiled in the kernel. Exiting\n");
		exit(1);
	}

	/* all the code is executed from a child proccess. If the child ends abnormally,
	there's still the parent for continue reading the /proc/supervesafb file and try to
	set another possible graphic mode */
	
	do {
		fread(buffer,7,1,salida); // read the command
		fclose(salida);
		if ((buffer[0]!=0)&&(buffer[2]<((unsigned char)maxconsole))) { // valid command and valid console
			pid_value=fork();
			if(pid_value==0) {

				ioperm(0, 1024, 1);
				
				if(buffer[0] & 0x01) // command for change the mode
						cambia_modo(buffer[1]);

				if(buffer[0] & 0x02) { // command for change a palette entry
					outb_p(buffer[3], dac_reg);
					outb_p(buffer[4], dac_val);
					outb_p(buffer[5], dac_val);
					outb_p(buffer[6], dac_val);
				}
				exit (0);
			} else
				if(pid_value>0)
					wait(NULL);
		}
		if(buffer[0]!=0)
			salida=fopen("/proc/supervesafb","r");
	} while(buffer[0]!=0); // if we read an invalid command, we have to exit
	
	return 0;
}
